SuperStrict

Rem
bbdoc: PostgreSQL Database Interface
End Rem
Module Pub.PostgreSQL

ModuleInfo "Name: BlitzMax PostgreSQL Module v1.00"
ModuleInfo "Description: BlitzMax module implementing client support for PostgreSQL"
ModuleInfo "License: freebie"
ModuleInfo "Authors: gman, original PostgreSQL imports by Chris Camacho"
ModuleInfo "History: 2006/1/18 GG - intial release v1.00"

Import Pub.ISQLDB
Import BRL.Blitz

Import "-lpq"

Public

Rem
bbdoc: Interface for a PostgreSQL database connection
EndRem
Type IPostgreSQLDB Extends ISQLDB Abstract
	
	Rem
	bbdoc: Gets the extra parameters passed in the connection string
	EndRem
	Method getParams:String() Abstract
	Rem
	bbdoc: Escapes a string to make it compatible with the PostgreSQL query engine
	EndRem
	Method escapeString:String(stmt:String) Abstract
	Rem
	bbdoc: Escapes a binary string to make it compatible with the PostgreSQL query engine
	EndRem
	Method escapeBinary:String(stmt:String) Abstract		
	Rem
	bbdoc: Unescapes a binary string
	EndRem
	Method unescapeBinary:String(stmt:String) Abstract
	
EndType

Rem
bbdoc: Gets an instance of the PostgreSQL database driver
about: Pass bConnect as True to connect to the server automatically.  Pass the returned value to SetISQLDBDriver to make this the global database driver.
EndRem
Function GetPostgreSQLDBDriver:ISQLDB(bConnect:Int,host:String,user:String,passwd:String,db:String=Null,port:Int=5432,extra_params:String="")
	Return PostgreSQLDB.create(bConnect,host,user,passwd,db,port,extra_params)
EndFunction

Private

Type PostgreSQLDB Extends IPostgreSQLDB

	Global _onendhandler:IOnEndResultEventHandler=Null
												
	Field _host:String
	Field _username:String
	Field _password:String
	Field _port:Int=5432
	Field _xtraparams:String=""
	Field _conn:Int
	Field _db:String
	Field _lastErrMsg:String
	Field _lastErr:Int
	Field _lastNativeErr:String

	Function create:IPostgreSQLDB(bConnect:Int,host:String,user:String,passwd:String,db:String=Null,port:Int=5432,extra_params:String="")
		Local retval:PostgreSQLDB=New PostgreSQLDB		
		retval.setHost(host)
		retval.setUsername(user)
		retval.setPassword(passwd)
		retval.setDB(db)
		retval.setPort(port)
		retval.setParams(extra_params)
		If bConnect Then retval.openConnection()
		Return retval
	EndFunction

	Function setOnResultHandler(handler:IOnResultEventHandler=Null)
		DebugLog("PostgreSQLDB does not support setOnResultHandler")
	EndFunction
	
	Function setOnEndResultHandler(handler:IOnEndResultEventHandler=Null)
		_onendhandler=handler
	EndFunction
						
	Method escapeString:String(stmt:String)
		Local retval:String=Null
		Local strTo:Byte Ptr=MemAlloc(stmt.length*2)
		Local length:Int=PQescapeString(strTo,stmt,stmt.length)
		If length>0 Then retval=String.FromBytes(strTo,length) Else FillError(ISQLDB.ERR_UNKNOWN,"Failed to escape string")
		MemFree(strTo)
		Return retval
	EndMethod
	
	Method escapeBinary:String(binary:String)
		Local retval:String=Null
		Local datalen:Int=0		
		Local data:Byte Ptr=PQescapeBytea(binary,binary.length,Varptr(datalen))
		If datalen>0 Then retval=String.FromBytes(data,datalen) Else FillError(ISQLDB.ERR_UNKNOWN,"Failed to escape binary data")
		PQfreemem(data)
		Return retval
	EndMethod
	
	Method unescapeBinary:String(binary:String)
		Local retval:String=Null
		Local datalen:Int=0
		Local data:Byte Ptr=PQunescapeBytea(binary,Varptr(datalen))
		If datalen>0 Then retval=String.FromBytes(data,datalen) Else FillError(ISQLDB.ERR_UNKNOWN,"Failed to escape binary data")
		PQfreemem(data)
		Return retval	
	EndMethod

	Method setHost(host:String)
		_host=host
	EndMethod 
	
	Method getHost:String()
		Return _host
	EndMethod
	
	Method setDB(database:String)
		_db=database
	EndMethod
	
	Method getDB:String()
		Return _db
	EndMethod
	
	Method setUsername(username:String)
		_username=username
	EndMethod
	
	Method getUsername:String()
		Return _username
	EndMethod
	
	Method setPassword(password:String)
		_password=password
	EndMethod
	
	Method getPassword:String()
		Return _password
	EndMethod
	
	Method setPort(port:Int=5432)
		_port=port
	EndMethod
	
	Method getPort:Int()
		Return _port
	EndMethod

	Method setParams(params:String="")
		_xtraparams=params
	EndMethod
	
	Method getParams:String()
		Return _xtraparams
	EndMethod
										
	Method BuildConnString:String()
		Local retval:String=""
		
		If getHost().length>0 Then retval:+"host="+getHost()+" "
		If getUsername().length>0 Then retval:+"user="+getUsername()+" "
		If getPassword().length>0 Then retval:+"password="+getPassword()+" "
		If getPort()>0 Then retval:+"port="+getPort()+" "
		If getParams().length>0 Then retval:+getParams()
		
		Return retval
	EndMethod
	
	' methods for working the connection
	Method openConnection:Int()
		_conn=PQconnectdb(BuildConnString())
		Local retval:Int=Not (PQstatus(_conn) = CONNECTION_BAD)
		
		If Not retval
			' error, but not loggable because no valid connection
			FillError(ISQLDB.ERR_INIT_FAIL,"Failed to intialize PostgreSQL client library")
		EndIf
		
		Return retval
	EndMethod
		
	Method closeConnection()
		If _conn Then PQfinish(_conn)
		_conn=0
	EndMethod
	
	Method Delete()
		closeConnection()
	EndMethod

	Method clearErrors()
		_lastErr=ISQLDB.ERR_NONE
		_lastErrMsg=Null
		_lastNativeErr=Null
	EndMethod

	Method FillError(errno:Int=ISQLDB.ERR_NONE,errmsg:String=Null)
		If errno=ISQLDB.ERR_NONE
			clearErrors()
			' see if there is an unknown error			
			If String(PQerrorMessage(_conn)).length>0 Then FillError(ISQLDB.ERR_UNKNOWN)
		Else
			_lastErr=errno
			If Not errmsg Then _lastErrMsg=PQerrorMessage(_conn) Else _lastErrMsg=errmsg
			_lastNativeErr=PQerrorMessage(_conn)			
		EndIf
	EndMethod
		
	Method getError:Int()
		Return _lastErr
	EndMethod
	
	Method getErrorMsg:String()
		Return _lastErrMsg
	EndMethod
	
	Method getNativeError:String()
		Return _lastNativeErr
	EndMethod
	
	Method getNativeHandle:Int()
		Return _conn
	EndMethod
		
	Method execute:ISQLResult(query:String)
		Local retval:PostgreSQLResult=Null

		clearErrors()

		Local res:Int = PQexec(_conn, query)

		If PQresultStatus(res) = PGRES_TUPLES_OK
										
			retval=New PostgreSQLResult
			retval._conn=Self
			retval._qry=query
			retval._result=res
			retval._recvd_rows=PQntuples(retval._result)
			
			' populate some meta info
			If retval._result
				' number of columns
				retval._colcnt=PQnfields(retval._result)								
				If _onendhandler Then _onendhandler.OnEndResult(retval)
			Else
				FillError(ISQLDB.ERR_RESULT_FAIL)
				Return Null		
			EndIf
		Else
			FillError(ISQLDB.ERR_QUERY_FAIL)			
		EndIf		
		
		Return retval
	EndMethod
	
	Method executeNonQuery:ISQLResult(stmt:String)
		Local retval:PostgreSQLResult=Null

		clearErrors()

		Local res:Int = PQexec(_conn, stmt)

		If PQresultStatus(res) <= PGRES_TUPLES_OK										
			retval=New PostgreSQLResult
			retval._conn=Self
			retval._qry=stmt
			retval._result=res
			retval._recvd_rows=String(PQcmdTuples(retval._result)).ToInt()
			If _onendhandler Then _onendhandler.OnEndResult(retval)
		Else
			FillError(ISQLDB.ERR_QUERY_FAIL)			
		EndIf		
		
		Return retval
	EndMethod
	
	Method executeScalar:String(query:String,field_nm:String=Null)
		Local retval:String=Null
		Local result:PostgreSQLResult=PostgreSQLResult(execute(query))
		
		' if the result failed, then just return to keep the error info intact
		If Not result Then Return retval

		If result.getNumRows()=0
			FillError(ISQLDB.ERR_NO_RESULTS,"There were no results returned by the query")
		Else			
			' if no field passed, return first row, first column
			If Not field_nm
				retval=result.getVal(0,0)
			Else
				retval=result.getValByName(0,field_nm)
			EndIf
			If _onendhandler Then _onendhandler.OnEndResult(result)
		EndIf
		Return retval	
	EndMethod
	
EndType

Type PostgreSQLResult Extends ISQLResult

	Field _conn:PostgreSQLDB=Null
	Field _qry:String=Null
	Field _result:Int
	Field _colcnt:Int=0
	Field _recvd_rows:Int=0
	Field _resulthandler:IOnResultEventHandler=Null
	Field _endresulthandler:IOnEndResultEventHandler=Null
	
	Method Delete()
		If _result Then PQclear(_result)
	EndMethod

	Method getComplete:Int()
		Return _result
	EndMethod

	Method getConnection:ISQLDB()
		Return _conn
	EndMethod
	
	Method getQuery:String()
		Return _qry
	EndMethod
	
	Method getError:Int()
		Return _conn.getError()
	EndMethod
	
	Method getErrorMsg:String()
		Return _conn.getErrorMsg()
	EndMethod
	
	Method getNativeError:String()
		Return _conn.getNativeError()
	EndMethod
	
	Method getSuccess:Int()
		Return (getError()=0)
	EndMethod
	
	Method getNumRows:Int()
		Return _recvd_rows
	EndMethod
	
	Method getNumCols:Int()
		Return _colcnt
	EndMethod
	
	Method getColIdx:Int(col:String)		
		Local retval:Int=PQfnumber(_result,col)
		If retval<0 Then _conn.FillError(ISQLDB.ERR_INVALID_COLUMN,"Invalid column '"+col+"'")		
		Return retval
	EndMethod
	
	Method getColName:String(col:Int)
		Local retval:String=Null
		
		If col>=0 And col<_colcnt
			retval=PQfname(_result,col)
			If Not retval Or retval.length=0 Then _conn.FillError(ISQLDB.ERR_INVALID_COLUMN,"Invalid column '"+col+"'")				
		Else
			_conn.FillError(ISQLDB.ERR_INVALID_COLUMN,"Invalid column '"+col+"'")
		EndIf
		
		Return retval		
	EndMethod
	
	Method getColWidth:Int(row:Int,col:Int)
		Local retval:Int=0
		
		If _recvd_rows>0 And row<_recvd_rows
			retval=PQgetlength(_result,row,col)
		Else
			If getError()=0 Then _conn.FillError(ISQLDB.ERR_INVALID_ROWCOL,"Invalid row/column ("+row+"/"+col+")")
		EndIf
		
		Return retval
	EndMethod
	
	Method getVal:String(row:Int,col:Int)
		Local retval:String=Null
		
		' make sure we have the row
		If _recvd_rows>0 And row<_recvd_rows
			retval=PQgetvalue(_result, row, col)
		Else
			If getError()=0 Then _conn.FillError(ISQLDB.ERR_INVALID_ROWCOL,"Invalid row/column ("+row+"/"+col+")")
		EndIf
		
		Return retval
	EndMethod
	
	Method getValByName:String(row:Int,col:String)
		Return getVal(row,getColIdx(col))
	EndMethod
	
	Method getIntVal:Int(row:Int,col:Int)
		Local val:String=getVal(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToInt()
	EndMethod
	
	Method getIntValByName:Int(row:Int,col:String)
		Local val:String=getValByName(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToInt()
	EndMethod
	
	Method getLongVal:Long(row:Int,col:Int)
		Local val:String=getVal(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToLong()
	EndMethod
	
	Method getLongValByName:Long(row:Int,col:String)
		Local val:String=getValByName(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToLong()
	EndMethod
	
	Method getFloatVal:Float(row:Int,col:Int)
		Local val:String=getVal(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToFloat()
	EndMethod

	Method getFloatValByName:Float(row:Int,col:String)
		Local val:String=getValByName(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToFloat()
	EndMethod
	
	Method getDoubleVal:Float(row:Int,col:Int)
		Local val:String=getVal(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToDouble()
	EndMethod
	
	Method getDoubleValByName:Float(row:Int,col:String)
		Local val:String=getValByName(row,col)
		If Not val Then RuntimeError "Invalid row/column ("+row+"/"+col+")"
		Return val.ToDouble()
	EndMethod
	
	Method getInsertId:String()
		Return String(PQoidStatus(_result))
	EndMethod

	Method getNativeHandle:Int()
		Return _result
	EndMethod
EndType

Public

Const CONNECTION_GOOD:Int=0
Const CONNECTION_BAD:Int=1
	
Const PGRES_EMPTY_QUERY:Int=0
Const PGRES_COMMAND_OK:Int=1
Const PGRES_TUPLES_OK:Int=2
Const PGRES_COPY_OUT:Int=3
Const PGRES_COPY_IN:Int=4
Const PGRES_BAD_RESPONSE:Int=5
Const PGRES_NONFATAL_ERROR:Int=6
Const PGRES_FATAL_ERROR:Int=7

Extern
	Rem
	bbdoc: Makes a New connection To the database server
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQconnectdb(s$z)
	Rem
	bbdoc: Returns the status of the connection
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQstatus:Int(conn:Int)
	Rem
	bbdoc: Returns the error message most recently generated by an operation on the connection
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQerrorMessage$z(conn:Int)
	Rem
	bbdoc: Submits a command to the server and waits for the result
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQexec:Int(conn:Int, query_string$z)
	Rem
	bbdoc: Closes an open PostgreSQL client connection
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQfinish(conn:Int)
	Rem
	bbdoc: Returns the number of rows (tuples) in the query result
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQntuples:Int(res:Int)
	Rem
	bbdoc: 
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQcmdTuples$z(res:Int)
	Rem
	bbdoc: Returns a string with the OID of the inserted row
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQoidStatus$z(res:Int)
	Rem
	bbdoc: Returns the result status of the command
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQresultStatus:Int(res:Int)
	Rem
	bbdoc: Frees the storage associated with a PGresult
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQclear(res:Int)
	Rem
	bbdoc: Returns a single field value of one row of a PGresult
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQgetvalue$z(res:Int,i:Int,n:Int)
	Rem
	bbdoc: Returns the column number associated with the given column name
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQfnumber:Int(res:Int,column_name$z)
	Rem
	bbdoc: Returns the column name associated with the given column number
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQfname$z(res:Int,column_number:Int)
	Rem
	bbdoc: Returns the number of columns (fields) in each row of the query result
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQnfields:Int(res:Int)
	Rem
	bbdoc: Returns the size in bytes of the column associated with the given column number
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQfsize:Int(res:Int,column_number:Int)
	Rem
	bbdoc: Tests a field for a null value
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQgetisnull:Int(res:Int,row_number:Int,column_number:Int)
	Rem
	bbdoc: Returns the actual length of a field value in bytes
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQgetlength:Int(res:Int,row_number:Int,column_number:Int)
	Rem
	bbdoc: Escapes a string for use within an SQL command
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQescapeString:Int(ptrTo:Byte Ptr,from$z,from_length:Int)
	Rem
	bbdoc: Escapes binary data for use within an SQL command with the type bytea
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQescapeBytea:Byte Ptr(from$z,from_length:Int,to_length:Int Ptr)
	Rem
	bbdoc: Converts an escaped string representation of binary data into binary data [mdash ] the reverse of PQescapeBytea
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQunescapeBytea:Byte Ptr(from$z,to_length:Int Ptr)
	Rem
	bbdoc: Frees the memory returned by PQescapeBytea and PQunescapeBytea
	about: See the PostgreSQL SDK manual for more information
	EndRem
	Function PQfreemem(dataptr:Byte Ptr) 
EndExtern

